import warnings

warnings.warn("pymatgen.matproj.snl has been moved to pymatgen.util.provenance."
              "This stub will be removed in pmg 2018.01.01.")

from pymatgen.util.provenance import *

